/*
**    TP ETI3 CPE Lyon
**    Copyright (C) 2008 Damien Rohmer
**
**    This program is free software: you can redistribute it and/or modify
**    it under the terms of the GNU General Public License as published by
**    the Free Software Foundation, either version 3 of the License, or
**    (at your option) any later version.
**
**   This program is distributed in the hope that it will be useful,
**    but WITHOUT ANY WARRANTY; without even the implied warranty of
**    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
**    GNU General Public License for more details.
**
**    You should have received a copy of the GNU General Public License
**    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef MC_V3D_VECTOR_HPP_
#define MC_V3D_VECTOR_HPP_



#include <MC_v3d.hpp>
#include <vector>
#include <set>
#include <map>
#include <list>
#include <algorithm>



namespace mesh_conv
{
    class MC_double_vector;
    class MC_int_vector;
    class MC_matrix;
    class MC_connectivity_index;
    class MC_mesh_index_vector;

    /// Container class for a vector of MC_v3d
    /** Internal structure is a std::vector<MV_v3d> */
    class MC_v3d_vector
    {


    protected:

        /** \brief internal storage */
        std::vector <MC_v3d> v;


    public:


        //*********************************************//
        //*********************************************//
        // CONSTRUCTORS
        //*********************************************//
        //*********************************************//

        /** \brief empty constructor, void vector*/
        MC_v3d_vector();
        /** \brief constructor of a zeros vector of a given size */
        MC_v3d_vector(const int& new_size);
        /** \brief direct constructor with a MC_v3d */
        MC_v3d_vector(const MC_v3d& v0);
        /** \brief direct constructor with a MC_v3d */
        MC_v3d_vector(const MC_v3d& v0,const MC_v3d& v1);
        /** \brief direct constructor with a MC_v3d */
        MC_v3d_vector(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2);
        /** \brief direct constructor with a MC_v3d */
        MC_v3d_vector(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3);
        /** \brief direct constructor with a MC_v3d */
        MC_v3d_vector(const MC_v3d& v0,const MC_v3d& v1,const MC_v3d& v2,const MC_v3d& v3,const MC_v3d& v4);
        /** \brief copy constructor */
        MC_v3d_vector(const MC_v3d_vector& vec);
        /** \brief direct constructor from a vector of MC_v3d */
        MC_v3d_vector(const std::vector <MC_v3d>& vec);
        /** \brief direct constructor from a set of MC_v3d */
        MC_v3d_vector(const std::set <MC_v3d,MC_v3d_less> set);
        /** \brief direct constructor from a map of MC_v3d */
        MC_v3d_vector(const std::map <MC_v3d,int,MC_v3d_less> map);
        /** \brief direct constructor from a vector of x,y and z coordinates */
        MC_v3d_vector(const MC_double_vector& x_vector,const MC_double_vector& y_vector,const MC_double_vector& z_vector);

        //************************************************//
        //************************************************//
        //Asserts
        //************************************************//
        //************************************************//

        /** \brief Assert the value stays within the bounds
            \return nothing, but stop the programm if it goes outside bounds
        */
        void assert_bounds(const int& u) const;



        //*********************************************//
        //*********************************************//
        // Attribute
        //*********************************************//
        //*********************************************//

        /** \brief get the size of the vector */
        int size() const;

        //*********************************************//
        //*********************************************//
        // Manipulate
        //*********************************************//
        //*********************************************//

        /** \brief clear the vector to size 0*/
        void clear();

        /** \brief resize the vector in deleting value of adding zeros
            \return the new value
        */
        MC_v3d_vector& resize(const int& new_size);

        /** \brief push back a R3 vector
            \return the new vector
        */
        MC_v3d_vector& add(const MC_v3d& x);
        /** \brief push back a vector of MC_v3d&
            \return the new vector
        */
        MC_v3d_vector& add(const MC_v3d_vector& v_x);
        /** \brief add a value at the end of the vertex
            \return the current vector (copy)
        */
        friend MC_v3d_vector operator<<(const MC_v3d_vector& vec,const MC_v3d& value);
        /** \brief add a vector of values at the end of the vertex
            \return the current vector (copy)
        */
        friend MC_v3d_vector operator<<(const MC_v3d_vector& vec,const MC_v3d_vector& value);


        /** \set set an indexed value to a certain component. Adjust the size automatically
            \return the new vector with the setted component
        */
        MC_v3d_vector& set(const int& k_index,const MC_v3d& x);
        /** \set set an indexed value to a certain component. Adjust the size automatically
            \return the new vector with the setted component

            k_index and x must have the same size (or x must be of size 1)
        */
        MC_v3d_vector& set(const mesh_conv::MC_int_vector& k_index,const MC_v3d_vector& x);


        /** \brief return the first element (error if the list is empty) */
        const MC_v3d& first() const;
        /** \brief return the first element (error if the list is empty) */
        MC_v3d& first();

        /** \brief return the last element (error if the list is empty) */
        const MC_v3d& last() const;
        /** \brief return the last element (error if the list is empty) */
        MC_v3d& last();

        /** \brief extract one of the dimension of the vector */
        MC_double_vector component(const int& k_dim) const;



        //************************************************//
        //************************************************//
        //Initialize value
        //************************************************//
        //************************************************//


        /** \brief build a vector of size new_size filled with zero */
        static MC_v3d_vector zeros(const int& new_size);


        //************************************************//
        //************************************************//
        //Math operator
        //************************************************//
        //************************************************//




        /** \brief change the opposite sign of the current vector (unary negation)
            \return a new vector with opposite sign
        */
        MC_v3d_vector operator-() const;


        /** \brief add a MC_v3d value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator+(const MC_v3d_vector& vec,const MC_v3d& to_add);
        /** \brief add a MC_v3d value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator+(const MC_v3d& to_add,const MC_v3d_vector& vec);
        /** \brief substract a MC_v3d value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator-(const MC_v3d_vector& vec,const MC_v3d& to_sub);

        /** \brief multiply a double value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator*(const MC_v3d_vector& vec,const double& to_mult);
        /** \brief multiply a double value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator*(const double& to_mult,const MC_v3d_vector& vec);
        /** \brief divide a double value to the vector
            \return the new vector
        */
        friend MC_v3d_vector operator/(const MC_v3d_vector& vec,const double& to_subdiv);



        /** \brief add a set of value to the vector
            \return the new vector

            to_add must be of size 1 or vec.size()
        */
        friend MC_v3d_vector operator+(const MC_v3d_vector& vec,const MC_v3d_vector& to_add);
        /** \brief substract a v3d value to the vector
            \return the new vector

            to_sub must be of size 1 or vec.size()
        */
        friend MC_v3d_vector operator-(const MC_v3d_vector& vec,const MC_v3d_vector& to_sub);



        /** \brief internal add a value */
        MC_v3d_vector& operator+=(const MC_v3d& to_add);
        /** \brief internal substraction */
        MC_v3d_vector& operator-=(const MC_v3d& to_sub);
        /** \brief internal multiplication */
        MC_v3d_vector& operator*=(const double& to_mult);
        /** \brief internal divide */
        MC_v3d_vector& operator/=(const double& to_subdiv);

        /** \brief internal add values

            to_add must be of size 1 or this->size()
        */
        MC_v3d_vector& operator+=(const MC_v3d_vector& to_add);
        /** \brief internal substraction

            to_add must be of size 1 or this->size()
        */
        MC_v3d_vector& operator-=(const MC_v3d_vector& to_sub);

        /** \brief do a summation over the MC_v3d_vector */
        static MC_v3d sum(const MC_v3d_vector& vec);

        /** \brief apply a matrix to the vector of MC_v3d */
        friend MC_v3d_vector operator*(const MC_matrix& M,const MC_v3d_vector& vec);
        /** \brief internal application of a matrix to the vector of MC_v3d (does the product x'=M*x) */
        MC_v3d_vector& operator*=(const MC_matrix& M);

        /** \brief apply a pointwise multiplication by a MC_double_vector
            \warning the MC_double_vector must be the same size as the vector or of size 1 */
        friend MC_v3d_vector operator*(const MC_double_vector w,const MC_v3d_vector& vec);
        /** \brief apply a pointwise multiplication by a MC_double_vector
            \warning the MC_double_vector must be the same size as the vector or of size 1 */
        friend MC_v3d_vector operator*(const MC_v3d_vector& vec,const MC_double_vector& w);
        /** \brief internal pointwise multiplication by a MC_double_vector
            \warning the MC_double_vector must be the same size as the vector or of size 1 */
        MC_v3d_vector& operator*=(const MC_double_vector& w);

        /** \brief return the norm a the vector independently */
        static MC_double_vector norm(const MC_v3d_vector& vec);

        /** \brief normalize each vector independently */
        MC_v3d_vector normalized() const;

        /** \brief internal scale the set of point by multipling componentwise with a v3d*/
        MC_v3d_vector& scale(const MC_v3d_vector& s);

        /** \brief homogeneous scale the point set to fit into [0,1]
            \return the new vector
            \return the applied scaling factor */
        std::pair <MC_v3d_vector,double> scaled_to_unit() const;

        /** \brief get the barycenter of the positions */
        MC_v3d barycenter() const;


        //*********************************************//
        //*********************************************//
        // get Operators
        //*********************************************//
        //*********************************************//

        /** \brief get operator  */
        const MC_v3d& operator()(const int& k_index) const ;
        /** \brief get operator  */
        MC_v3d& operator()(const int& k_index) ;
        /** \brief get operator  */
        const MC_v3d& operator[](const int& k_index) const ;
        /** \brief get operator  */
        MC_v3d& operator[](const int& k_index) ;
        
        /** \brief get a subset of the vector */
        MC_v3d_vector operator()(const MC_int_vector& index) const; 
        /** \brief get a subset of the vector */
        MC_v3d_vector operator[](const MC_int_vector& index) const;

        /** \brief get pointer */
        const MC_v3d* pointer() const;
        /** \brief get pointer
            \warning not protected
        */
        MC_v3d* pointer_unprotected();

        //************************************************//
        //************************************************//
        //change type
        //************************************************//
        //************************************************//

        /** \brief convert to std::map (coords is key and store the position as value) */
        std::map <MC_v3d,int,MC_v3d_less> to_map() const;
        /** \brief convert to std::set */
        std::set <MC_v3d,MC_v3d_less> to_set() const;
        /** \brief convert to list in the same order */
        std::list <MC_v3d> to_list() const;


        //*********************************************//
        //*********************************************//
        // Bounding box
        //*********************************************//
        //*********************************************//

        /** \brief get the extreme position min and max
            \return min
            \return max
        */
        std::pair <MC_v3d,MC_v3d> bounding_box_elements() const;

        /** \brief return the length in x,y,z coordinates of the bounding box */
        MC_v3d bounding_box_length() const;


        //*********************************************//
        //*********************************************//
        // INPUT OUTPUT
        //*********************************************//
        //*********************************************//

        /** \brief write to ostream the list (u0 u1 u2 ... )*/
        friend std::ostream& operator<<(std::ostream& output,const MC_v3d_vector& in);

        /** \brief write to a stream in a transfering format (each coordinate separated by a space) */
        std::ostream& export_stream(std::ostream& output) const;
        /** \brief read from a stream in a transfering format (each coordinate separated by a space) */
        MC_v3d_vector& read_stream(std::istream& input);

    private:
    };


    /// Comparator class for set and map of MC_v3d_vector
    class MC_v3d_vector_less
    {
    public:
        bool operator()(const MC_v3d_vector& v0,const MC_v3d_vector& v1) const
        {
            MC_v3d_less L;
            if(v0.size()<v1.size())
                return true;
            else if(v0.size()>v1.size())
                return false;
            else //equal size compare by values
            {
                int N=v0.size();
                for(int k=0;k<N;++k)
                {
                    if(L(v0[k],v1[k])==true)
                        return true;
                    else if(L(v1[k],v0[k])==true)
                        return false;
                }
                //complete equality
                return false;
            }
        }
    };

}

#endif
